@@ -89,9 +89,12 @@ class RWKV(MyModule):
             free_slots = 0
             for i in range(len(s)):
                 si = s[i]
-                if si[1] == 'fp32': si[1] = torch.float
-                elif si[1] == 'fp16': si[1] = torch.float16
-                elif si[1] == 'bf16': si[1] = torch.bfloat16
+                si1 = si[1]
+                if si1.startswith('fp32'): si[1] = [torch.float]
+                elif si1.startswith('fp16'): si[1] = [torch.float16]
+                elif si1.startswith('bf16'): si[1] = [torch.bfloat16]
+                if si1.endswith('i8'): si[1] += [torch.uint8]
+                else: si[1] += [si[1][0]]
                 if len(si) > 2:
                     ss = si[2]
                     assert ss.startswith('*')
@@ -123,9 +126,9 @@ class RWKV(MyModule):
             for i in range(len(s)):
                 ss = s[i]
                 if i != stream_i:
-                    print(f'* {ss[0]} {ss[1]}, store {plan[i]} layers')
+                    print(f'* {ss[0]} {str(ss[1]).replace("torch.","")}, store {plan[i]} layers')
                 else:
-                    print(f'* {ss[0]} {ss[1]}, store {plan[i]-stream_count} layers, stream {stream_count} layers')
+                    print(f'* {ss[0]} {str(ss[1]).replace("torch.","")}, store {plan[i]-stream_count} layers, stream {stream_count} layers')
                 plan[i] += (0 if i == 0 else plan[i-1])
             self.strategy = [None] * (args.n_layer + 1)
             strategy = self.strategy
@@ -134,12 +137,13 @@ class RWKV(MyModule):
                     if n < plan[i]:
                         strategy[n] = types.SimpleNamespace()
                         strategy[n].device = s[i][0]
-                        strategy[n].dtype = s[i][1]
+                        strategy[n].atype = s[i][1][0]
+                        strategy[n].wtype = s[i][1][1]
                         strategy[n].stream = False
                         if i == stream_i and n >= (plan[i] - stream_count):
                             strategy[n].stream = True
                         break
-                print(f"{n}-{strategy[n].device}-{str(strategy[n].dtype).replace('torch.','')}{'-stream' if strategy[n].stream else ''}",end=' ')
+                print(f"{n}-{strategy[n].device}-{str(strategy[n].atype).replace('torch.','')}-{str(strategy[n].wtype).replace('torch.','')}{'-stream' if strategy[n].stream else ''}",end=' ')
             print()
 
             # Load weights
@@ -151,8 +155,15 @@ class RWKV(MyModule):
                     layer_id = args.n_layer
                 dd = strategy[layer_id]
                 DEVICE = dd.device
-                DTYPE = dd.dtype
-                
+                ATYPE = dd.atype
+                WTYPE = dd.wtype
+
+                if self.RESCALE_LAYER > 0:
+                    if 'att.output.weight' in x:
+                        w[x] = w[x] / (2 ** int(layer_id // self.RESCALE_LAYER))
+                    if 'ffn.value.weight' in x:
+                        w[x] = w[x] / (2 ** int(layer_id // self.RESCALE_LAYER))
+
                 if '.time_' in x:
                     w[x] = w[x].squeeze()
                 if 'key.weight' in x or 'value.weight' in x or 'receptance.weight' in x or 'output.weight' in x or 'head.weight' in x:
@@ -163,23 +174,43 @@ class RWKV(MyModule):
                 elif '.time_first' in x: # need fp32 for this
                     w[x] = w[x].float()
                 else:
-                    w[x] = w[x].to(dtype=DTYPE)
-
-                if self.RESCALE_LAYER > 0:
-                    if 'att.output.weight' in x:
-                        w[x] = w[x] / (2 ** int(layer_id // self.RESCALE_LAYER))
-                    if 'ffn.value.weight' in x:
-                        w[x] = w[x] / (2 ** int(layer_id // self.RESCALE_LAYER))
+                    if (len(w[x].shape) == 2) and ('emb' not in x):
+                        if WTYPE != torch.uint8:
+                            w[x] = w[x].to(dtype=WTYPE)
+                        else:
+                            w[x] = w[x].float()
+                            w[x+'_mx'] = torch.amin(w[x], dim=0)
+                            w[x+'_rx'] = torch.amax(w[x], dim=0) - w[x+'_mx']
+                            w[x] = (w[x] - w[x+'_mx']) / w[x+'_rx']
+                            w[x+'_my'] = torch.amin(w[x], dim=1)
+                            w[x+'_ry'] = torch.amax(w[x], dim=1) - w[x+'_my']
+                            w[x+'_my'] = w[x+'_my'].unsqueeze(1)
+                            w[x+'_ry'] = w[x+'_ry'].unsqueeze(1)
+                            w[x] = (w[x] - w[x+'_my']) / w[x+'_ry']
+                            w[x] = torch.round(w[x] * 255.0).to(dtype=torch.uint8)
+                            w[x+'_mx'] = w[x+'_mx'].to(dtype=ATYPE)
+                            w[x+'_rx'] = w[x+'_rx'].to(dtype=ATYPE)
+                            w[x+'_my'] = (w[x+'_my'] * 255.0).to(dtype=ATYPE)
+                            w[x+'_ry'] = w[x+'_ry'].to(dtype=ATYPE)
+                    else:
+                        w[x] = w[x].to(dtype=ATYPE)
                 
                 if 'emb.' in x:
                     pass
-                elif (dd.stream) and (('key.weight' in x) or ('value.weight' in x) or ('receptance.weight' in x) or ('output.weight' in x)):
+                elif (dd.stream) and (x.endswith('key.weight') or x.endswith('value.weight') or x.endswith('receptance.weight') or x.endswith('output.weight')):
                     try:
                         w[x] = w[x].pin_memory() # if you see "CUDA error: out of memory" here, that's out of CPU RAM, not VRAM. Get more RAM :)
                     except:
                         print('Note: You are running out of RAM. Get more CPU RAM. Now this will run much slower.')
                 elif DEVICE != 'cpu':
                     w[x] = w[x].to(device=DEVICE)
+                    try:
+                        w[x+'_mx'] = w[x+'_mx'].to(device=DEVICE)
+                        w[x+'_rx'] = w[x+'_rx'].to(device=DEVICE)
+                        w[x+'_my'] = w[x+'_my'].to(device=DEVICE)
+                        w[x+'_ry'] = w[x+'_ry'].to(device=DEVICE)
+                    except:
+                        pass
 
                 if 'ffn.value.weight' in x:
                     gc.collect()
@@ -196,8 +227,8 @@ class RWKV(MyModule):
                         print('\n', end = '')
                         print_need_newline = False
                     dt = str(w[x].dtype).replace('torch.', '')
-                    dt = dt.replace('float32', 'fp32').replace('bfloat16', 'bf16').replace('float16', 'fp16')
-                    print(x.ljust(32), dt, str(w[x].device).rjust(8), shape, ' (pinned)' if w[x].is_pinned() else '')
+                    dt = dt.replace('float32', 'f32').replace('bfloat16', 'bf16').replace('float16', 'f16').replace('uint8', 'i8')
+                    print(x.ljust(32), dt.rjust(4), str(w[x].device).rjust(8), shape, ' (pinned)' if w[x].is_pinned() else '')
                 else:
                     print_need_newline = True
                     print('.', end = '', flush = True)
@@ -205,7 +236,17 @@ class RWKV(MyModule):
             gc.collect()
             if 'cuda' in args.strategy_string:
                 torch.cuda.empty_cache()
-        
+
+    def get_w(self, x, dtype):
+        w = self.w
+        if w[x].dtype != torch.uint8:
+            return w[x]
+        return self.uint8_to_type(w[x].to(dtype=dtype), w[x+'_mx'], w[x+'_my'], w[x+'_rx'], w[x+'_ry'])
+
+    @MyFunction
+    def uint8_to_type(self, x, mx, my, rx, ry):
+        return (x * ry + my) * rx / 255.0 + mx
+
     @MyFunction
     def ffn_one(self, x, sx, ln_w, ln_b, k_mix, r_mix, kw, vw, rw):
         xx = F.layer_norm(x, (x.shape[-1],), weight=ln_w, bias=ln_b)
@@ -244,7 +285,7 @@ class RWKV(MyModule):
         p = torch.maximum(pp, ww)
         e1 = torch.exp(pp - p)
         e2 = torch.exp(ww - p)
-        wkv = ((e1 * aa + e2 * v) / (e1 * bb + e2)).to(dtype=r.dtype)
+        wkv = ((e1 * aa + e2 * v) / (e1 * bb + e2)).to(dtype=x.dtype)
         ww = t_decay + pp
         p = torch.maximum(ww, k)
         e1 = torch.exp(ww - p)
@@ -273,7 +314,7 @@ class RWKV(MyModule):
             p = torch.maximum(pp, ww)
             e1 = torch.exp(pp - p)
             e2 = torch.exp(ww - p)
-            sx[t] = ((e1 * aa + e2 * vv) / (e1 * bb + e2)).to(dtype=r.dtype)
+            sx[t] = ((e1 * aa + e2 * vv) / (e1 * bb + e2)).to(dtype=x.dtype)
             ww = t_decay + pp
             p = torch.maximum(ww, kk)
             e1 = torch.exp(ww - p)
@@ -313,12 +354,12 @@ class RWKV(MyModule):
                 for i in range(args.n_layer): # state: 0=att_xx 1=att_aa 2=att_bb 3=att_pp 4=ffn_xx
                     dd = self.strategy[i]
                     dev = dd.device
-                    dtype = dd.dtype
-                    state[i*5+0] = torch.zeros(args.n_embd, dtype=dtype, requires_grad=False, device=dev)
+                    atype = dd.atype
+                    state[i*5+0] = torch.zeros(args.n_embd, dtype=atype, requires_grad=False, device=dev)
                     state[i*5+1] = torch.zeros(args.n_embd, dtype=torch.float, requires_grad=False, device=dev)
                     state[i*5+2] = torch.zeros(args.n_embd, dtype=torch.float, requires_grad=False, device=dev)
                     state[i*5+3] = torch.zeros(args.n_embd, dtype=torch.float, requires_grad=False, device=dev) - 1e30
-                    state[i*5+4] = torch.zeros(args.n_embd, dtype=dtype, requires_grad=False, device=dev)
+                    state[i*5+4] = torch.zeros(args.n_embd, dtype=atype, requires_grad=False, device=dev)
 
             seq_mode = len(tokens) > 1
 
@@ -330,7 +371,8 @@ class RWKV(MyModule):
                 ffn = f'blocks.{i}.ffn.'
                 dd = self.strategy[i]
                 dev = dd.device
-                dtype = dd.dtype
+                atype = dd.atype
+                wtype = dd.wtype
                 if seq_mode:
                     if 'cuda' in str(dev) and os.environ["RWKV_CUDA_ON"] == '1':
                         ATT = self.cuda_att_seq
@@ -341,60 +383,58 @@ class RWKV(MyModule):
                     ATT = self.att_one
                     FFN = self.ffn_one
 
-                x = x.to(dtype=dtype, device=dev)
+                x = x.to(dtype=atype, device=dev)
+
+                kw = self.get_w(f'{att}key.weight', atype)
+                vw = self.get_w(f'{att}value.weight', atype)
+                rw = self.get_w(f'{att}receptance.weight', atype)
+                ow = self.get_w(f'{att}output.weight', atype)
                 if dd.stream:
-                    kw = w[f'{att}key.weight'].to(device=dev, non_blocking=True)
-                    vw = w[f'{att}value.weight'].to(device=dev, non_blocking=True)
-                    rw = w[f'{att}receptance.weight'].to(device=dev, non_blocking=True)
-                    ow = w[f'{att}output.weight'].to(device=dev, non_blocking=True)
-                    x, state[i*5+0], state[i*5+1], state[i*5+2], state[i*5+3] = ATT(
-                        x, sx=state[i*5+0], aa=state[i*5+1], bb=state[i*5+2], pp=state[i*5+3],
-                        ln_w=w[f'{bbb}ln1.weight'], ln_b=w[f'{bbb}ln1.bias'],
-                        k_mix=w[f'{att}time_mix_k'], v_mix=w[f'{att}time_mix_v'], r_mix=w[f'{att}time_mix_r'],
-                        t_decay = w[f'{att}time_decay'], t_first = w[f'{att}time_first'],
-                        kw=kw, vw=vw, rw=rw, ow=ow)
+                    kw = kw.to(device=dev, non_blocking=True)
+                    vw = vw.to(device=dev, non_blocking=True)
+                    rw = rw.to(device=dev, non_blocking=True)
+                    ow = ow.to(device=dev, non_blocking=True)
+                x, state[i*5+0], state[i*5+1], state[i*5+2], state[i*5+3] = ATT(
+                    x, sx=state[i*5+0], aa=state[i*5+1], bb=state[i*5+2], pp=state[i*5+3],
+                    ln_w=w[f'{bbb}ln1.weight'], ln_b=w[f'{bbb}ln1.bias'],
+                    k_mix=w[f'{att}time_mix_k'], v_mix=w[f'{att}time_mix_v'], r_mix=w[f'{att}time_mix_r'],
+                    t_decay = w[f'{att}time_decay'], t_first = w[f'{att}time_first'],
+                    kw=kw, vw=vw, rw=rw, ow=ow)
+                if wtype == torch.uint8 or dd.stream:
                     del kw
                     del vw
                     del rw
                     del ow
-                else:
-                    x, state[i*5+0], state[i*5+1], state[i*5+2], state[i*5+3] = ATT(
-                        x, sx=state[i*5+0], aa=state[i*5+1], bb=state[i*5+2], pp=state[i*5+3],
-                        ln_w=w[f'{bbb}ln1.weight'], ln_b=w[f'{bbb}ln1.bias'],
-                        k_mix=w[f'{att}time_mix_k'], v_mix=w[f'{att}time_mix_v'], r_mix=w[f'{att}time_mix_r'],
-                        t_decay = w[f'{att}time_decay'], t_first = w[f'{att}time_first'],
-                        kw=w[f'{att}key.weight'],
-                        vw=w[f'{att}value.weight'],
-                        rw=w[f'{att}receptance.weight'],
-                        ow=w[f'{att}output.weight'])
+
+                kw = self.get_w(f'{ffn}key.weight', atype)
+                vw = self.get_w(f'{ffn}value.weight', atype)
+                rw = self.get_w(f'{ffn}receptance.weight', atype)
                 if dd.stream:
-                    kw = w[f'{ffn}key.weight'].to(device=dev, non_blocking=True)
-                    vw = w[f'{ffn}value.weight'].to(device=dev, non_blocking=True)
-                    rw = w[f'{ffn}receptance.weight'].to(device=dev, non_blocking=True)
-                    x, state[i*5+4] = FFN(
-                        x, sx=state[i*5+4],
-                        ln_w=w[f'{bbb}ln2.weight'], ln_b=w[f'{bbb}ln2.bias'],
-                        k_mix=w[f'{ffn}time_mix_k'], r_mix=w[f'{ffn}time_mix_r'],
-                        kw=kw, vw=vw, rw=rw)
+                    kw = kw.to(device=dev, non_blocking=True)
+                    vw = vw.to(device=dev, non_blocking=True)
+                    rw = rw.to(device=dev, non_blocking=True)
+                x, state[i*5+4] = FFN(
+                    x, sx=state[i*5+4],
+                    ln_w=w[f'{bbb}ln2.weight'], ln_b=w[f'{bbb}ln2.bias'],
+                    k_mix=w[f'{ffn}time_mix_k'], r_mix=w[f'{ffn}time_mix_r'],
+                    kw=kw, vw=vw, rw=rw)
+                if wtype == torch.uint8 or dd.stream:                
                     del kw
                     del vw
                     del rw
-                else:
-                    x, state[i*5+4] = FFN(
-                        x, sx=state[i*5+4],
-                        ln_w=w[f'{bbb}ln2.weight'], ln_b=w[f'{bbb}ln2.bias'],
-                        k_mix=w[f'{ffn}time_mix_k'], r_mix=w[f'{ffn}time_mix_r'],
-                        kw=w[f'{ffn}key.weight'],
-                        vw=w[f'{ffn}value.weight'],
-                        rw=w[f'{ffn}receptance.weight'])
 
                 if self.RESCALE_LAYER > 0:
                     if (i+1) % self.RESCALE_LAYER == 0:
                         x = x / 2
-            
+
+            dd = self.strategy[args.n_layer]
             x = x[-1,:] if (seq_mode and (not full_output)) else x
-            x = x.to(dtype=self.strategy[args.n_layer].dtype, device=self.strategy[args.n_layer].device)
+            x = x.to(dtype=dd.atype, device=dd.device)
+            
             x = F.layer_norm(x, (args.n_embd,), weight=w['ln_out.weight'], bias=w['ln_out.bias'])
-            x = x @ w['head.weight']
+            if w['head.weight'].dtype != torch.uint8:
+                x = x @ w['head.weight']
+            else:
+                x = x @ self.get_w('head.weight', dd.atype)
 
             return x.float(), state
